begin
  require "numru/narray"
rescue LoadError
  require "narray"
end
require "tmpdir"

module NumRu
  module NArrayExt

    @@tmpnum = 0

    def self.c_loop(*arys,&block)
      @@tmpnum += 1
      na = -1
      args = []
      @@loop = Array.new
      ars = arys.map do |ary|
        na += 1
        vn = "v#{na}"
        args.push vn
        NArrayExt::Ary.new(ary.shape, ary.rank, ary.typecode, [:ary, vn], @@loop)
      end
      @@body = ""
      self.module_exec(*ars,&block)
      tmpnum = 0
      func_name = "rb_narray_ext_#{@@tmpnum}"
      code = <<EOF
#include "ruby.h"
#include "narray.h"

VALUE
#{func_name}(VALUE self, VALUE rb_#{args.join(", VALUE rb_")}) {
EOF
      ars.each_with_index do |ary,i|
        ctype = ary.ctype
        code << "  #{ctype} v#{i} = NA_PTR_TYPE(rb_v#{i}, #{ctype});\n"
      end
      code << @@body
      code << <<EOF
  return Qnil;
}

void
Init_narray_ext_#{@@tmpnum}()
{
  VALUE mNumRu = rb_define_module("NumRu");
  VALUE mNArrayExt = rb_define_module_under(mNumRu, "NArrayExt");
  rb_define_module_function(mNArrayExt, "c_func_#{@@tmpnum}", #{func_name}, #{ars.length});
}
EOF
#      tmpdir = "tmp"
#      Dir.mkdir(tmpdir) unless File.exist?(tmpdir)
      Dir.mktmpdir(nil, ".") do |tmpdir|
        Dir.chdir(tmpdir) do |dir|
          fname = "narray_ext_#{@@tmpnum}.c"
          File.open(fname, "w") do |file|
            file.print code
          end
          File.open("extconf.rb", "w") do |file|
            file.print extconf(@@tmpnum)
          end
          unless system("ruby extconf.rb > log && make >> log")
            print File.read("log"), "\n"
            print File.read(fname), "\n"
            raise("compile error")
          end
          require "./narray_ext_#{@@tmpnum}.so"
          NumRu::NArrayExt.send("c_func_#{@@tmpnum}", *arys)
        end
      end

      return nil
    end

    def self.loop(min, max)
      i = @@loop.length
      idx = NArrayExt::Index.new(i, min, max)
      @@body << "  "*(i+1)
      @@body << "for (int i#{i}=#{min}; i#{i}<=#{max}; i#{i}++) {\n"

      @@loop.push Array.new
      yield(idx)
      offset = "  "*(i+2)
      @@loop.pop.each do |ex|
        @@body << offset
        @@body << (String===ex ? ex : ex.to_c)
      end
      @@body << "  "*(i+1)+"}\n"
    end

    def self.extconf(tmpnum)
      src = <<EOF
require "mkmf"

gem_path = nil
begin
  require "rubygems"
  if Gem::Specification.respond_to?(:find_by_name)
    if spec = ( Gem::Specification.find_by_name("numru-narray") || Gem::Specification.find_by_name("narray") )
      gem_path = spec.full_gem_path
    end
  else
    if (spec = (Gem.source_index.find_name("numru-narray") || Gem.source_index.find_name("narray")) ).any?
      gem_path = spec.full_gem_path
    end
  end
  gem_path = File.join(gem_path, "ext", "numru", "narray")
rescue LoadError
  dir_config("narray", RbConfig::CONFIG["sitearchdir"])
end

unless find_header("narray.h", gem_path)
  find_header("narray.h", File.join(gem_path,"src"))
end

create_makefile("narray_ext_#{tmpnum}")
EOF
    end



    class Index

      def initialize(idx,min,max,shift=nil)
        @idx = idx
        @min = min
        @max = max
        @shift = shift
      end

      def +(i)
        case i
        when Fixnum
          NArrayExt::Index.new(@idx,@min,@max,i)
        else
          raise ArgumentError, "invalid argument"
        end
      end

      def -(i)
        case i
        when Fixnum
          NArrayExt::Index.new(@idx,@min,@max,-i)
        else
          raise ArgumentError, "invalid argument"
        end
      end

      def to_s
        if @shift.nil?
          "i#{@idx}"
        elsif @shift > 0
          "i#{@idx}+#{@shift}"
        else # < 0
          "i#{@idx}#{@shift}"
        end
      end
      alias :inspect :to_s

      def idx
        @idx
      end
      def min
        @shift ? [@min, @min+@shift].min : @min
      end
      def max
        @shift ? [@max, @max+@shift].max : @max
      end

    end

    class Ary

      @@ntype2ctype = {NArray::BYTE => "u_int8_t*",
                       NArray::SINT => "int16_t*",
                       NArray::LINT => "int32_t*",
                       NArray::SFLOAT => "float*",
                       NArray::DFLOAT => "double*",
                       NArray::SCOMPLEX => "scomplex*",
                       NArray::DCOMPLEX => "dcomplex*"}
      if NArray.const_defined?(:LLINT)
        @@ntype2ctype[NArray::LLINT] = "int64_t*"
      end

      def initialize(shape, rank, type, ops, exec=nil)
        @shape = shape
        @rank = rank
        @type = type
        @ops = ops
        @exec = exec
        if @type==NArray::SCOMPLEX || @type==NArray::DCOMPLEX
          raise "complex is not supported at this moment"
        end
      end

      def [](*idx)
        unless idx.length == @rank
          raise "number of idx != rank"
        end
        self.class.new(@shape, 0, @type, [:slice, @ops, slice(*idx)])
      end

      def []=(*arg)
        idx = arg[0..-2]
        other = arg[-1]
        unless idx.length == @rank
          raise "number of idx != rank"
        end
        ops = [:slice, @ops, slice(*idx)]
        case other
        when Numeric
          @exec[-1].push self.class.new(@shape, @rank, @type, [:set, ops, [:int,other]])
        when self.class
          @exec[-1].push self.class.new(@shape, @rank, @type, [:set, ops, other.ops])
        else
          raise ArgumentError, "invalid argument"
        end
        return nil
      end

      def +(other)
        arithmetic(other, :add)
      end

      def -(other)
        arithmetic(other, :sub)
      end

      def *(other)
        arithmetic(other, :mul)
      end

      def /(other)
        arithmetic(other, :div)
      end

      def to_c
        get_str(@ops) + ";\n"
      end

      def ctype
        @@ntype2ctype[@type]
      end

      protected
      def ops
        @ops
      end

      @@twoop = {:add => "+", :sub => "-", :mul => "*", :div => "/", :set => "="}

      private
      def get_str(obj)
        op = obj[0]
        obj = obj[1..-1]
        case op
        when :slice
          obj, idx = obj
          obj = get_str(obj)
          if @rank==1
            sidx = idx[0].to_s
          else
            sidx = idx[-1].to_s
            (@rank-1).times do |i|
              sidx = "(#{sidx})*#{@shape[-i-2]}+(#{idx[-i-2].to_s})"
            end
          end
          return "#{obj}[#{sidx}]"
        when :add, :sub, :mul, :div
          o1, o2 = obj
          o1 = get_str(o1)
          o2 = get_str(o2)
          return "( #{o1} #{@@twoop[op]} #{o2} )"
        when :set
          o1, o2 = obj
          o1 = get_str(o1)
          o2 = get_str(o2)
          return "#{o1} #{@@twoop[op]} #{o2}"
        when :int, :ary
          return obj[0].to_s
        else
          raise "unexpected value: #{op} (#{op.class})"
        end
      end

      def arithmetic(other, op)
        unless @rank==0
          raise "slice first"
        end
        case other
        when Numeric
          return self.class.new(@shape, 0, @type, [op, @ops, [:int,other]])
        when self.class
          return self.class.new(@shape, 0, @type, [op, @ops, other.ops])
        else
          raise ArgumentError, "invalid argument: #{other} (#{other.class})"
        end
      end

      def slice(*idx)
        case @rank
        when 0
          raise "rank is zero"
        when 1
          idx = [idx]
        else
          unless Array===idx && idx.length == @rank
            raise ArgumentError, "number of index must be equal to rank"
          end
        end
        idx.each_with_index do |id,i|
          case id
          when NArrayExt::Index
            if id.min < 0 || id.max > @shape[i]-1
              raise ArgumentError, "out of boundary"
            end
          when Fixnum
            idx[i] = id + @shape[i] if id < 0
            if id > @shape[i]-1
              raise ArgumentError, "out of boundary"
            end
          else
            raise ArgumentError, "index is invalid: #{i.class}"
          end
        end
        return idx
      end

    end # class Ary

  end # module NArrayExt

end # module NumRu
